/* i386 only */

#define N_TXTLEN(x) \
	((N_MAGIC(x)==ZMAGIC ? sizeof (struct exec) : 0) + (x).a_text)
#define N_DATLEN(x) ((x).a_data)
#define N_BSSLEN(x) ((x).a_bss)

#define	N_TXTOFF(x) \
	(N_MAGIC(x)==ZMAGIC ? 0 : sizeof (struct exec))
#define N_DATOFF(x) \
	(N_TXTOFF(x) + N_TXTLEN(x))

#define N_TXTADDR(x) 0x10000
#define N_DATADDR(x) (N_TXTADDR(x) + N_TXTLEN (x))
#define N_BSSADDR(x) (N_DATADDR(x) + N_DATLEN (x))

#include "a.out.gnu.h"
