/* 
   Copyright (C) 1994 Free Software Foundation

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "trans.h"

/* Implement fshelp_transboot_clean; the port library clean function
   for transboot ports.  */
void
fshelp_transboot_clean (void *arg)
{
  struct transboot *tb = arg;
  
  mach_port_deallocate (mach_task_self (), tb->node);

  mutex_lock (&tb->link->lock);
  if (tb->link->control == MACH_PORT_NULL)
    {
      tb->link->error = EDIED;
      tb->link->starting = 0;
      condition_broadcast (&tb->link->initwait);
    }
  mutex_unlock (&tb->link->lock);
}
