/* Hash table definitions
   Copyright (C) 1993 Free Software Foundation

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#define N_HASH_BUCKETS 511

struct port_hash_elt
{
  struct port_info *pi;
  struct port_hash_elt *next, **prevp;
};

extern spin_lock_t _libports_hash_lock;
extern struct port_hash_elt *_libports_port_hash_buckets[];
