/* 
   Copyright (C) 1993, 1994 Free Software Foundation

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "ports.h"
#include <assert.h>

mach_port_t _libports_request_portset;
spin_lock_t _libports_portrefcntlock = SPIN_LOCK_INITIALIZER;

/* Initialize the ports library.  */
void
_libports_initialize ()
{
  int err;
  err = mach_port_allocate (mach_task_self (), MACH_PORT_RIGHT_PORT_SET, 
			    &_libports_request_portset);
  assert (!err);
}

