/*
   Copyright (C) 1993, 1994 Free Software Foundation

This file is part of the GNU Hurd.

The GNU Hurd is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

The GNU Hurd is distributed in the hope that it will be useful, 
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the GNU Hurd; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Written by Michael I. Bushnell.  */

#include "ports.h"

/* Lookup PORT, allocate a user reference to the associated port structure 
   and return it.  If PORT was not created by ports_allocate_port, 
   then return 0.  */
void *
ports_get_port (mach_port_t port)
{
  struct port_info *pi;
  
  pi = _libports_hash_lookup (port);
  if (pi)
    {
      spin_lock (&_libports_portrefcntlock);
      pi->refcnt++;
      spin_unlock (&_libports_portrefcntlock);
    }
  return pi;
}
